#include <WiFi.h>
#include <MQTT.h>

// Soil Moisture sensor pin (replace with your pin number)
#define SOIL_MOISTURE_PIN 34

int led1_pin = 2;
int soilMoistureValue = 0;

#define WIFI_SSID             "IOT@SPAK"
#define WIFI_PASSWORD         "i0t@SP4K"
#define MQTT_HOST             "broker.hivemq.com"
#define MQTT_PREFIX_TOPIC     "csm3313_umt/group03"
#define MQTT_PUBLISH_TOPIC1   "/soil_moisture"
#define MQTT_SUBSCRIBE_TOPIC1 "/led01"

WiFiClient net;
MQTTClient mqtt(1024);
unsigned long lastMillis = 0;

void connectToWiFi() {
  Serial.print("Connecting to Wi-Fi '" + String(WIFI_SSID) + "' ...");

  // Connect to WiFi
  WiFi.begin(WIFI_SSID, WIFI_PASSWORD);

  // while wifi not connected yet, print '.'
  // then after it connected, get out of the loop
  Serial.println("Connecting to WiFi...");
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  // print a new line, then print WiFi connected and the IP address
  Serial.println("");
  Serial.println("WiFi connected");
  // Print the IP address
  Serial.println(WiFi.localIP());
}

void messageReceived(String topic, String payload) {
  Serial.println("Incoming Status from topic " + topic + " -> " + payload);

  // check if the topic equals MQTT_SUBSCRIBE_TOPIC1
  if (topic == (String(MQTT_PREFIX_TOPIC) + String(MQTT_SUBSCRIBE_TOPIC1))) {
    if (payload == "1") {
      digitalWrite(led1_pin, HIGH);
      Serial.println("LED1 turned ON");
    } else if (payload == "0") {
      digitalWrite(led1_pin, LOW);
      Serial.println("LED1 turned OFF");
    }
  } else {
    Serial.println("Command does not match.");
  }
}

void connectToMqttBroker() {
  Serial.print("Connecting to '" + String(WIFI_SSID) + "' ...");

  mqtt.begin(MQTT_HOST, net);
  mqtt.onMessage(messageReceived);

  String uniqueString = String(WIFI_SSID) + "-" + String(random(1, 98)) + String(random(99, 999));
  char uniqueClientID[uniqueString.length() + 1];

  uniqueString.toCharArray(uniqueClientID, uniqueString.length() + 1);

  while (!mqtt.connect(uniqueClientID)) {
    Serial.print(".");
    delay(500);
  }

  Serial.println(" connected!");

  Serial.println("Subscribe to: " + String(MQTT_PREFIX_TOPIC) + String(MQTT_SUBSCRIBE_TOPIC1));
  mqtt.subscribe(String(MQTT_PREFIX_TOPIC) + String(MQTT_SUBSCRIBE_TOPIC1));
}

void setup(void) {
  Serial.begin(115200);
  pinMode(led1_pin, OUTPUT);
  digitalWrite(led1_pin, LOW);

  connectToWiFi();
  connectToMqttBroker();
  Serial.println();

  //Enable timer wakeup from deep sleep
  //esp_sleep_enable_time_wakeup(300000);
}

void loop() {
  mqtt.loop();
  delay(10); // <- fixes some issues with WiFi stability

  if (WiFi.status() != WL_CONNECTED) {
    connectToWiFi();
  }

  if (!mqtt.connected()) {
    connectToMqttBroker();
  }

  if (millis() - lastMillis > 900000) { // publish every 5 minutes
    lastMillis = millis();
    
    // Read soil moisture sensor
    soilMoistureValue = analogRead(SOIL_MOISTURE_PIN);
    int moisturePercentage = map(soilMoistureValue, 0, 4095, 100, 0);

    Serial.print("Soil Moisture: ");
    Serial.print(moisturePercentage);
    Serial.println("%");

    // Determine soil moisture status category
    String soilMoistureStatus;
    if (moisturePercentage <= 20) {
      soilMoistureStatus = "Very Dry";
    } else if (moisturePercentage <= 40) {
      soilMoistureStatus = "Dry";
    } else if (moisturePercentage <= 60) {
      soilMoistureStatus = "Wet";
    } else {
      soilMoistureStatus = "Very Wet";
    }

    Serial.println("Soil Moisture Status: " + soilMoistureStatus);

    // Convert sensor readings to String and publish to broker
    String strSoilMoisture = String(moisturePercentage) + "," + soilMoistureStatus;

    Serial.println("Publish to topic: " + String(MQTT_PREFIX_TOPIC) + String(MQTT_PUBLISH_TOPIC1));
    Serial.println("Data: " + strSoilMoisture);
    mqtt.publish(String(MQTT_PREFIX_TOPIC) + String(MQTT_PUBLISH_TOPIC1), strSoilMoisture);

    // ... (existing code)
  }
  Serial.println("..");
  delay(500);
}
